"""
This module contains some processing functions that while not belonging
to a specific class, require them to already be loaded and DISSTANS to
be initialized.

For general helper functions, see :mod:`~disstans.tools`.
"""

import numpy as np
import pandas as pd
import warnings
from functools import wraps
from sklearn.decomposition import PCA, FastICA
from scipy.signal import find_peaks
from tqdm import tqdm
from warnings import warn
from pandas.api.indexers import BaseIndexer

from .config import defaults
from .timeseries import Timeseries
from .compiled import selectpair
from .tools import Timedelta, parallelize, tvec_to_numpycol, date2decyear
from .models import Polynomial


def unwrap_dict_and_ts(func):
    """
    A wrapper decorator that aims at simplifying the coding of processing functions.
    Ideally, a new function that doesn't need to know if its input is a
    :class:`~disstans.timeseries.Timeseries`, :class:`~pandas.DataFrame`,
    :class:`~numpy.ndarray` or a dictionary containing them, should not need to reimplement
    a check and conversion for all of these because they just represent a data
    array of some form. So, by providing this function decorator, a wrapped function
    only needs to be able to work for a single array (plus some optional keyword arguments).
    The wrapping will extract the data of the input types and convert the returned
    array from ``func`` into the original format.

    Example
    -------
    A basic function of the form::

        def func(in_array, **kw_args):
            # do some things
            return out_array

    that takes and returns a NumPy array can be wrapped as follows
    to be able to also take and return all the other data forms::

        @unwrap_dict_and_ts
        def func(in_array, **kw_args):
            # do some things
            return out_array
    """
    @wraps(func)
    def wrapper(data, *args, **kw_args):
        if not isinstance(data, dict):
            data = {'ts': data}
            was_dict = False
        else:
            was_dict = True
        out = {}
        additional_output = {}
        # loop over components
        for comp, ts in data.items():
            if isinstance(ts, Timeseries):
                array = ts.data.values
            elif isinstance(ts, pd.DataFrame):
                array = ts.values
            elif isinstance(ts, np.ndarray):
                array = ts
            else:
                raise TypeError(f"Cannot unwrap object of type {type(ts)}.")
            func_output = func(array, *args, **kw_args)
            if isinstance(func_output, tuple):
                result = func_output[0]
                try:
                    additional_output[comp] = func_output[1:]
                except IndexError:
                    additional_output[comp] = None
            else:
                result = func_output
                additional_output[comp] = None
            # save results
            if isinstance(ts, Timeseries):
                out[comp] = ts.copy(only_data=True, src=func.__name__)
                out[comp].data = result
            elif isinstance(ts, pd.DataFrame):
                out[comp] = ts.copy()
                out[comp].values = result
            else:
                out[comp] = result
        if all([elem is None for elem in additional_output.values()]):
            has_additional_output = False
        else:
            has_additional_output = True
        if not was_dict:
            out = out['ts']
            additional_output = additional_output['ts']
        if has_additional_output:
            return out, additional_output
        else:
            return out
    return wrapper


class ExpandingRollingIndexer(BaseIndexer):
    """
    Indexer class used in the pandas rolling calculations (e.g., rolling median).
    This custom indexer behaves like a regular rolling window, except at the
    start and end of the timeseries, where the window slowly grows or shrinks,
    respectively. Specifically, at the first time step, the window is only one
    element wide. At the second step, it is three elements wide; at the third,
    five elements, and so forth, until the window has the defined window size.
    As the window approaches the end of the timeseries, it starts to shrink again
    until it is one element wide at the last time step.

    The goal of this kind of rolling window is to reduce the influence of late
    (early) observations when looking at the first (last) timesteps (respectively).

    For more information, see the pandas documentation about `custom window rolling
    <https://pandas.pydata.org/docs/user_guide/window.html#custom-window-rolling>`_.
    """
    def get_window_bounds(self, num_values, min_periods, center, closed):
        """
        This is the function that needs to be implemented for the
        :class:`~pandas.api.indexers.BaseIndexer` class to be used for all
        rolling pandas operations. It shouldn't be called manually.

        The parameters ``min_periods``, ``center``, and ``closed`` have to
        be included to match expected pandas behavior, but this function
        does not implement their potential features.

        Parameters
        ----------
        num_values : int
            Total number of values in the array to be rolled through.

        Returns
        -------
        start : np.ndarray
            Start indices of individual windows.
        end  : np.ndarray
            End indices (exclusive) of individual windows.
        """
        # window_size is set by BaseIndexer.__init__
        assert self.window_size % 2 == 1, \
            f"'window_size' must be odd, got {self.window_size}."
        self.window_size = self.window_size
        # initialize index arrays
        start = np.empty(num_values, dtype=int)
        end = np.empty_like(start)
        # left and right boundaries for the boundary intervals
        max_ix_start = min(max(1, num_values // 2 + 1), self.window_size // 2)
        max_ix_end = min(max(1, num_values // 2), self.window_size // 2)
        # starting interval
        start[:max_ix_start] = np.zeros(max_ix_start)
        end[:max_ix_start] = 2*np.arange(max_ix_start)
        # ending interval
        start[-max_ix_end:] = num_values - 1 - 2 * np.arange(max_ix_end - 1, -1, -1)
        end[-max_ix_end:] = (num_values - 1) * np.ones(max_ix_end)
        # middle, regular interval
        start[self.window_size//2:-(self.window_size//2)] = \
            np.arange(0, num_values - self.window_size + 1)
        end[self.window_size//2:-(self.window_size//2)] = \
            np.arange(self.window_size - 1, num_values)
        # return
        return start, end + 1


@unwrap_dict_and_ts
def median(array, kernel_size):
    """
    Computes the rolling median filter column-wise. Missing observations (NaNs) are
    ignored during the median calculation, but missing observations are not imputed
    from the rolling median (i.e., a NaN value remains a NaN value, but does not
    affect surrounding values).

    Parameters
    ----------
    array : numpy.ndarray
        2D input array (can contain NaNs).
        Wrapped by :func:`~disstans.processing.unwrap_dict_and_ts` to also accept
        :class:`~disstans.timeseries.Timeseries`, :class:`~pandas.DataFrame` and
        dictionaries of them as input.
    kernel_size : int
        Kernel size (length of moving window to compute the median over).
        Has to be an odd number.

    Returns
    -------
    filtered : numpy.ndarray
        2D filtered array (may still contain NaNs).
    """
    # make sure the array is 2D even if it's only a single column
    num_obs = array.shape[0]
    array = array.reshape(num_obs, 1 if array.ndim == 1 else -1)
    # the indexer object will tell pandas over which windows to calculate the median
    indexer = ExpandingRollingIndexer(window_size=kernel_size)
    # now we calculate the rolling median (this is compiled-optimized by pandas)
    filtered = pd.DataFrame(array).rolling(indexer, min_periods=1, axis=0).median().values
    # add NaNs again to where they were initially
    filtered[np.isnan(array)] = np.NaN
    return filtered


@unwrap_dict_and_ts
def decompose(array, method, num_components=1, return_sources=False):
    r"""
    Decomposes the input signal into different components using PCA or ICA.

    Parameters
    ----------
    array : numpy.ndarray
        Input array of shape :math:`(\text{num_observations},\text{n_stations})`
        (can contain NaNs).
        Wrapped by :func:`~disstans.processing.unwrap_dict_and_ts` to also accept
        :class:`~disstans.timeseries.Timeseries`, :class:`~pandas.DataFrame` and
        dictionaries of them as input (i.e. the output of
        :meth:`~disstans.network.Network.export_network_ts`).
    method : str
        Method to use to decompose the array. Possible values are ``'pca'`` and ``'ica'``:
        ``'pca'`` uses :class:`~sklearn.decomposition.PCA` (motivated by [dong06]_), whereas
        ``'ica'`` uses :class:`~sklearn.decomposition.FastICA` (motivated by [huang12]_).
    num_components : int, optional
        Number of components to estimate. Defaults to ``1``. If ``None``, all are used.
    return_sources : bool, optional
        If ``True``, return not only the best-fit model, but also the sources
        themselves in space and time. Defaults to ``False``.

    Returns
    -------
    model : numpy.ndarray
        Best-fit model with shape :math:`(\text{num_observations},\text{n_stations})`.
    temporal : numpy.ndarray
        Only if ``return_sources=True``: Temporal source with shape
        :math:`(\text{num_observations},\text{num_components})`.
    spatial : numpy.ndarray
        Only if ``return_sources=True``: Spatial source with shape
        :math:`(\text{num_components},\text{n_stations})`.


    References
    ----------

    .. [dong06] Dong, D., Fang, P., Bock, Y., Webb, F., Prawirodirdjo, L.,
       Kedar, S., & Jamason, P. (2006). *Spatiotemporal filtering using principal component
       analysis and Karhunen-Loeve expansion approaches for regional GPS network analysis*.
       Journal of Geophysical Research: Solid Earth, 111(B3).
       doi:`10.1029/2005JB003806 <https://doi.org/10.1029/2005JB003806>`_.
    .. [huang12] Huang, D. W., Dai, W. J., & Luo, F. X. (2012).
       *ICA Spatiotemporal Filtering Method and Its Application in GPS Deformation Monitoring*.
       Applied Mechanics and Materials, 204–208, 2806–2812.
       doi:`10.4028/www.scientific.net/AMM.204-208.2806
       <http://dx.doi.org/10.4028/www.scientific.net/AMM.204-208.2806>`_.
    """
    # ignore all only-NaN columns
    array_nanind = np.isnan(array)
    finite_cols = np.nonzero(~array_nanind.all(axis=0))[0]
    nan_cols = np.nonzero(array_nanind.all(axis=0))[0]
    array = array[:, finite_cols]
    # fill NaNs with white Gaussian noise
    array_nanmean = np.nanmean(array, axis=0)
    array_nansd = np.nanstd(array, axis=0)
    array_nanind = np.isnan(array)
    rng = np.random.default_rng()
    for icol in range(array.shape[1]):
        array[array_nanind[:, icol], icol] = array_nanmean[icol] + \
            array_nansd[icol] * rng.normal(size=array_nanind[:, icol].sum())
    # decompose using the specified solver
    if method == 'pca':
        decomposer = PCA(n_components=num_components, whiten=True)
    elif method == 'ica':
        decomposer = FastICA(n_components=num_components, whiten="unit-variance")
    else:
        raise NotImplementedError("Cannot estimate the common mode error "
                                  f"using the '{method}' method.")
    # extract temporal component and build model
    temporal = decomposer.fit_transform(array)
    model = decomposer.inverse_transform(temporal)
    # reduce to where original timeseries were not NaNs and return
    model[array_nanind] = np.NaN
    if nan_cols != []:
        newmod = np.empty((temporal.shape[0], len(finite_cols) + len(nan_cols)))
        newmod[:, finite_cols] = model
        newmod[:, nan_cols] = np.NaN
        model = newmod
    if return_sources:
        spatial = decomposer.components_
        if nan_cols != []:
            newspat = np.empty((spatial.shape[0], len(finite_cols) + len(nan_cols)))
            newspat[:, finite_cols] = spatial
            newspat[:, nan_cols] = np.NaN
            spatial = newspat
        return model, temporal, spatial
    else:
        return model


def clean(station, ts_in, reference, ts_out=None,
          clean_kw_args={}, reference_callable_args={}):
    """
    Function operating on a single station's timeseries to clean it from outliers,
    and mask it out if the data is not good enough. The criteria are set by
    :attr:`~disstans.config.defaults` but can be overriden by providing ``clean_kw_args``.
    The criteria are:

    - ``'min_obs'``: Minimum number of observations the timeseries has to contain.
    - ``'std_outlier'``: Classify as an outlier any observation that is this many
      standard deviations away from the reference.
    - ``'iqr_outlier'``: Classify as an outlier any observation that is this many
      inter-quartile ranges (IQR, difference between the 25th and 75th percentile)
      away from the reference's 25th-75th percentile range.
    - ``'std_thresh'``: After the removal of outliers, the maximum standard deviation
      that the residual between reference and input timeseries is allowed to have.
    - ``'min_clean_obs'``: After the removal of outliers, the minimum number of
      observations the timeseries has to contain.

    Parameters
    ----------
    station : disstans.station.Station
        Station to operate on.
    ts_in : str
        Description of the timeseries to clean.
    reference : str, disstans.timeseries.Timeseries, function
        Reference timeseries.
        If string, checks for a timeseries with that description in the ``station``.
        If a :class:`~disstans.timeseries.Timeseries` instance, use it directly.
        If a function, the reference timeseries will be calculated as
        ``t_ref = reference(ts_in, **reference_callable_args)``.
    ts_out : str, optional
        If provided, duplicate ``ts_in`` to a new timeseries ``ts_out``
        and clean the copy (to preserve the raw timeseries).
    clean_kw_args : dict, optional
        Override the default cleaning criteria in :attr:`~disstans.config.defaults`,
        see the explanations above.
    reference_callable_args : dict, optional
        If ``reference`` is a function, ``reference_callable_args`` can be used
        to pass additional keyword arguments to the former when calculating
        the reference timeseries.

    Warning
    -------
    By default, this function operates *in-place*. If you don't wish to overwrite
    the raw input timeseries, specify ``ts_out``.
    """
    clean_settings = defaults["clean"].copy()
    clean_settings.update(clean_kw_args)
    # check if timeseries is present
    if ts_in not in station.timeseries:
        warn(f"Could not find timeseries '{ts_in}' in station {station.name}.",
             category=RuntimeWarning, stacklevel=2)
        return
    # check if we're modifying in-place or copying
    if ts_out is None:
        ts = station[ts_in]
    else:
        ts = station[ts_in].copy(only_data=True, src='clean')
    # check if we have a reference time series or need to calculate one
    # in the latter case, the input is name of function to call
    if not (isinstance(reference, Timeseries)
            or isinstance(reference, str)
            or callable(reference)):
        raise TypeError("'reference' has to either be a Timeseries, the name of one, "
                        f"or a function, got {type(reference)}.")
    if isinstance(reference, Timeseries):
        ts_ref = reference
    elif isinstance(reference, str):
        # get reference time series
        ts_ref = station[reference]
    elif callable(reference):
        ts_ref = reference(ts, **reference_callable_args)
    # check that both timeseries have the same data columns
    if not ts_ref.data_cols == ts.data_cols:
        raise ValueError("Reference time series has to have the same data columns as "
                         f"input time series, but got {ts_ref.data_cols} and {ts.data_cols}.")
    for dcol in ts.data_cols:
        # check for minimum number of observations
        if ts[dcol].count() < clean_settings["min_obs"]:
            ts.mask_out(dcol)
            continue
        # compute residuals
        if (clean_settings["std_outlier"] is not None) \
           or (clean_settings["iqr_outlier"] is not None) \
           or (clean_settings["std_thresh"] is not None):
            residual = ts[dcol].values - ts_ref[dcol].values
            sd = np.nanstd(residual)
        # check for and remove outliers
        if (clean_settings["std_outlier"] is not None) \
           or (clean_settings["iqr_outlier"] is not None):
            mask = ~np.isnan(residual)
            mask_copy = mask.copy()
            if clean_settings["std_outlier"] is not None:
                mask[mask_copy] &= (np.abs(residual[mask_copy])
                                    > clean_settings["std_outlier"] * sd)
            if clean_settings["iqr_outlier"] is not None:
                q1 = np.nanpercentile(residual, 25)
                q3 = np.nanpercentile(residual, 75)
                iqr = q3 - q1
                mask[mask_copy] &= np.logical_or(residual[mask_copy]
                                                 < q1 - clean_settings["iqr_outlier"] * iqr,
                                                 residual[mask_copy]
                                                 > q3 + clean_settings["iqr_outlier"] * iqr)
            ts[dcol][mask] = np.NaN
            residual = ts[dcol].values - ts_ref[dcol].values
            sd = np.nanstd(residual)
        # check for minimum number of clean observations
        if ts[dcol].count() < clean_settings["min_clean_obs"]:
            ts.mask_out(dcol)
            continue
        # check if total standard deviation is still too large
        if (clean_settings["std_thresh"] is not None) and (sd > clean_settings["std_thresh"]):
            ts.mask_out(dcol)
    # if we made a copy, add it to the station, otherwise we're already done
    if ts_out is not None:
        station.add_timeseries(ts_out, ts)


def midas(ts, steps=None, tolerance=0.001):
    """
    This function performs the MIDAS estimate as described by [blewitt16]_.
    It is adapted from the Fortran code provided by the author (see
    :func:`~disstans.compiled.selectpair` for more details and original copyright).

    MIDAS returns the median estimate of secular (constant) velocities in all data
    components using data pairs spanning exactly one year. By not including pairs
    crossing known step epochs and using a fixed period, the influence of unmodeled
    jumps and seasonal variations can be minimized. At the end, an empirical estimate
    of the velocities' uncertainty is calculated as well.

    Parameters
    ----------
    ts : disstans.timeseries.Timeseries
        Timeseries to perform the MIDAS algorithm on.
    steps : pandas.Series, pandas.DatetimeIndex, optional
        If given, a pandas Series or Index of step times, across which no pairs
        should be formed.
    tolerance : float, optional
        Tolerance when enforcing the one-year period of pairs (in 365.25-days-long years`).

    Returns
    -------
    mdl : disstans.models.Polynomial
        Fitted polynomial (offset & constant velocity) model.
    res : disstans.timeseries.Timeseries
        Residual timeseries.
    stats : dict
        Fittings statistics computed along the way.
        ``'num_epochs'``, ``'num_used'``, ``'num_pairs'``, and ``'nstep'`` are the number of
        epochs in ``ts``, the number of epochs used in the velocity pairs, the number of pairs
        formed, and the number of included steps, respectively. ``'frac_removed'`` and
        ``'sd_velpairs'`` are the fraction of removed pairs (because of velocity pairs more
        than two standard deviations away from their medians) and the estimated standard
        deviation of the velocity pairs, respectively, and for each component.

    References
    ----------

    .. [blewitt16] Blewitt, G., Kreemer, C., Hammond, W. C., & Gazeaux, J. (2016).
       *MIDAS robust trend estimator for accurate GPS station velocities without step detection.*
       Journal of Geophysical Research: Solid Earth, 121(3), 2054–2068.
       doi:`10.1002/2015JB012552 <https://doi.org/10.1002/2015JB012552>`_
    """
    # extract timeseries data, adjust zero-crossing to first epoch
    x_off = ts.data.iloc[0, :].values.reshape(1, -1)
    x = ts.data.values - x_off
    # convert timeseries index and step times to decimal years
    t = date2decyear(ts.time)
    tstep = np.zeros(1)  # need at least one entry for selectpair
    if steps is None:
        tstep_back = tstep
    else:
        steps_decyear = date2decyear(steps)
        tstep_back = np.concatenate([-steps_decyear[::-1], tstep])
        tstep = np.concatenate([steps_decyear, tstep])
    # get forward and backwards time pairs
    num_pairs, ip = selectpair(t, tstep, tolerance)
    if num_pairs >= ip.shape[1]:
        warn(f"Forward call to selectpair returned maximum number of pairs ({num_pairs} "
             f"with a maximum of {ip.shape[1]}). Consider re-compiling DISSTANS "
             "with a higher maxn constant in compiled.f90.", stacklevel=2)
    nb, ipb = selectpair(-t[::-1], tstep_back, tolerance)
    if nb >= ipb.shape[1]:
        warn(f"Backward call to selectpair returned maximum number of pairs ({nb} "
             f"with a maximum of {ipb.shape[1]}). Consider re-compiling DISSTANS "
             "with a higher maxn constant in compiled.f90.", stacklevel=2)
    if num_pairs + nb < 10:
        warn(f"Only found {num_pairs} forward and {nb} backward pairs; solution will be bad.",
             stacklevel=2)
    # convert backward indices to forward ones
    ipb = t.size - ipb[[1, 0], :nb]
    # combine the two index collections, and make them more readable
    ip = np.concatenate([ip[:, :num_pairs] - 1, ipb], axis=1)
    ip_from, ip_to = ip[0, :], ip[1, :]
    num_pairs += nb
    # calculate number of points used in pairing
    num_used = np.unique(ip).size
    # compute velocity for all pairs
    v = (x[ip_to, :] - x[ip_from, :]) / (t[ip_to] - t[ip_from]).reshape(-1, 1)
    # median of the velocities
    v50 = np.nanmedian(v, axis=0, keepdims=True)
    # absolute deviation from the median
    d = np.abs(v - v50)
    # median absolute deviation (MAD)
    d50 = np.nanmedian(d, axis=0, keepdims=True)
    # estimated standard deviation of velocities
    # (based on theoretical factor 1.4826)
    sd_velpairs = 1.4826 * d50
    # delete velocities more than 2 s.d. from MAD
    v[d >= 2*sd_velpairs] = np.NaN
    num_kept = (~np.isnan(v)).sum(axis=0, keepdims=True)
    # recompute median, absolute deviation, MAD, and estimated s.d.
    v50 = np.nanmedian(v, axis=0, keepdims=True)
    d = np.abs(v - v50)
    d50 = np.nanmedian(d, axis=0, keepdims=True)
    sd_velpairs = 1.4826 * d50
    # Standard errors for the median velocity
    # Multiply by theoretical factor of sqrt(pi/2) = 1.2533
    # Divide number of data by 4 since we use coordinate data a nominal 4 times
    # Also scale standard errors by ad hoc factor of 3 to be realistic
    sv = 1.2533 * sd_velpairs / np.sqrt(num_kept / 4) * 3
    # compute intercepts
    r = x - v50 * (t.reshape(-1, 1) - t[0])
    x50 = np.nanmedian(r, axis=0, keepdims=True)
    # compute residuals
    r -= x50
    # fraction of pairs removed
    frac_removed = (num_pairs - num_kept) / num_pairs
    # return offet & velocity as linear Polynomial model
    # intercept uncertainty is "perfect" since it wasn't estimated at all,
    # it's just relative to a reference time
    mdl = Polynomial(order=1, t_reference=ts.time[0], time_unit="Y")
    mdl.read_parameters(parameters=np.concatenate([x50 + x_off, v50], axis=0),
                        covariances=np.concatenate([np.zeros_like(x50), sv], axis=0))
    # return residual as a Timeseries
    res = Timeseries.from_array(timevector=ts.time, data=r, src="midas",
                                data_unit=ts.data_unit,
                                data_cols=[f"{dcol}_midasres" for dcol in ts.data_cols])
    # all other stats are returned in a dictionary
    stats = {"num_epochs": t.size, "num_used": num_used, "num_pairs": num_pairs,
             "frac_removed": frac_removed.ravel(), "sd_velpairs": sd_velpairs.ravel(),
             "nstep": tstep.size - 1}
    # return
    return mdl, res, stats


class StepDetector():
    r"""
    This class implements a step detector based on the Akaike Information Criterion (AIC).

    A window is moved over the input data, and two linear models are fit in the
    method :meth:`~search`: one containing only a linear polynomial, and one containing an
    additional step in the middle of the window. Then, using the AIC, the relative
    probabilities are calculated, and saved for each timestep.

    In the final step, one can threshold these relative probabilities with the method
    :meth:`~steps`, and look for local maxima, which will correspond to probable steps.

    If the class is constructed with ``kernel_size``, ``x`` and ``y`` passed, it automatically
    calls its method :meth:`~search`, otherwise, :meth:`~search` needs to be called manually.
    Running the method again with a different ``kernel_size``, ``x`` or ``y`` will overwrite
    previous results.

    Parameters
    ----------
    kernel_size : int, optional
        Window size of the detector. Must be odd.
    kernel_size_min : int, optional
        Minimum window size of the detector (for edges). Must be smaller than or equal
        to ``kernel_size``. Defaults to ``0``.
    x : numpy.ndarray, optional
        Input array of shape :math:`(\text{num_observations},)`.
        Should not contain NaNs.
    y : numpy.ndarray, optional
        Input array of shape :math:`(\text{num_observations}, \text{num_components})`.
        Can contain NaNs.
    """
    def __init__(self, kernel_size=None, kernel_size_min=0, x=None, y=None):
        self.kernel_size = kernel_size
        self.kernel_size_min = kernel_size_min
        if (x is not None) and (y is not None) and (kernel_size is not None):
            self.search(x, y, kernel_size)

    @property
    def kernel_size(self):
        """ Kernel (window) size of the detector. """
        if self._kernel_size is None:
            raise ValueError("'kernel_size' has not yet been set.")
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, kernel_size):
        if kernel_size is not None:
            assert isinstance(kernel_size, int) and (kernel_size % 2 == 1), \
                f"'kernel_size' must be an odd integer or None, got {kernel_size}."
        self._kernel_size = kernel_size

    @property
    def kernel_size_min(self):
        """ Minimum kernel (window) size of the detector. """
        if self._kernel_size_min is None:
            raise ValueError("'kernel_size_min' has not yet been set.")
        return self._kernel_size_min

    @kernel_size_min.setter
    def kernel_size_min(self, kernel_size_min):
        if kernel_size_min is not None:
            assert kernel_size_min <= self.kernel_size, "'kernel_size_min' must be smaller " + \
                f"or equal to 'kernel_size', but {kernel_size_min} > {self.kernel_size}."
        self._kernel_size_min = kernel_size_min

    @staticmethod
    def AIC_c(rss, n, K):
        r"""
        Calculates the Akaike Information Criterion for small samples for Least Squares
        regression results. Implementation is based on [burnhamanderson02]_ (ch. 2).

        Parameters
        ----------
        rss : float
            Residual sum of squares.
        n : int
            Number of samples.
        K : int
            Degrees of freedom. If the Least Squares model has :math:`\text{num_parameters}`
            parameters (including the mean), then the degrees of freedom are
            :math:`K = \text{num_parameters} + 1`

        Returns
        -------
        float
            The Small Sample AIC for Least Squares :math:`\text{AIC}_c`.

        References
        ----------

        .. [burnhamanderson02] (2002) *Information and Likelihood Theory:
           A Basis for Model Selection and Inference.* In: Burnham K.P., Anderson D.R. (eds)
           Model Selection and Multimodel Inference. Springer, New York, NY.
           doi:`10.1007/978-0-387-22456-5_2 <https://doi.org/10.1007/978-0-387-22456-5_2>`_.
        """
        # input check
        if n - K - 1 <= 0:
            # can't return meaningful statistic, hypothesis unlikely
            return np.NaN
        # calculate AIC for LS
        AIC = n * np.log(rss / n) + 2*K
        # apply small sample size correction
        correction = 2 * K * (K + 1) / (n - K - 1)
        return AIC + correction

    @staticmethod
    def test_single(xwindow, ywindow, valid=None, maxdel=10):
        r"""
        For a single window (of arbitrary, but odd length), perform the AIC hypothesis test
        whether a step is likely present (H1) or not (H0) in the ``y`` data given
        ``x`` coordinates.

        Parameters
        ----------
        xwindow : numpy.ndarray
            Time array of shape :math:`(\text{num_window},)`.
            Should not contain NaNs.
        ywindow : numpy.ndarray
            Data array of shape :math:`(\text{num_window},)`.
            Should not contain NaNs.
        valid : numpy.ndarray, optional
            Mask array of the data of shape :math:`(\text{num_window},)`,
            with ``1`` where the ``ywindow`` is finite (not NaN or infinity).
            If not passed to the function, it is calculated internally, which will slow
            down the computation.
        maxdel : float, optional
            Difference in AIC that should be considered not significantly better.
            (Refers to :math:`\Delta_i = \text{AIC}_{c,i} - \text{AIC}_{c,\text{min}}`.)

        Returns
        -------
        int
            Best hypothesis (``0`` for no step, ``1`` for step).
        float
            If H1 is the best hypothesis (and suffices ``maxdel``), its relative probability,
            otherwise the relative probability of H0 (which therefore can be ``0`` if H0 is also
            the best hypothesis in general).
        tuple
            A 2-tuple of the two mean-squared residuals of the H0 and H1 hypotheses,
            respectively. Assuming the test is unbiased, this is the residual's variance.
            Is ``NaN`` in an element if the least-squares model did not converge.

        See Also
        --------
        AIC_c : For more information about the AIC hypothesis test.
        """
        # do some checks
        assert xwindow.shape[0] == ywindow.shape[0], \
            "'xwindow' and 'ywindow' have to have the same length in the first dimensions, " \
            f"got {xwindow.shape} and {ywindow.shape}."
        assert (xwindow.shape[0] % 2 == 1), \
            "'xwindow' and 'ywindow' must have an odd number of entries, " \
            f"got {xwindow.shape[0]}."
        if valid is None:
            valid = np.isfinite(ywindow)
        else:
            assert ywindow.shape == valid.shape, \
                "'ywindow' and 'valid' have to have the same shape, " \
                f"got {ywindow.shape} and {valid.shape}."
        # get number of valid observations
        i_mid = int(xwindow.shape[0] // 2)
        n_pre = valid[:i_mid].sum()
        n_post = valid[i_mid:].sum()
        n_total = n_pre + n_post
        # return with 0, 0 if we will not be able to get an estimate because of not enough data
        if (n_pre < 2) or (n_post < 2):
            return 0, 0, (np.NaN, np.NaN)
        xfinite = xwindow[valid]
        yfinite = ywindow[valid]
        # build mapping matrix for model H1
        G1 = np.zeros((n_total, 3))
        # first column is mean
        G1[:, 0] = 1
        # second column is slope
        G1[:, 1] = xfinite - xfinite[0]
        # third column is the additional step
        G1[n_pre:, 2] = 1
        # without the step it's just the first two columns
        G0 = G1[:, :2]
        # fit for H1 first
        # (since if that one doesn't converge, we have to go with H0 anyway)
        try:
            rss1 = float(np.linalg.lstsq(G1, yfinite, rcond=None)[1])
        except np.linalg.LinAlgError:
            return 0, 0, (np.NaN, np.NaN)
        # now we can fit for H0, and again just go with that if there is no solution
        try:
            rss0 = float(np.linalg.lstsq(G0, yfinite, rcond=None)[1])
        except np.linalg.LinAlgError:
            return 0, 0, (np.NaN, rss1/n_total)
        # now that both models produce results, let's get the AIC_c values
        # we'll again return the H0 if not both models have a valid AIC_c value
        aic = [StepDetector.AIC_c(rss, n_total, dof) for (rss, dof)
               in zip([rss0, rss1], [3, 4])]
        if np.isnan(aic).sum() > 0:
            return 0, 0, (rss0/n_total, rss1/n_total)
        # let's check the difference between the two as a measure of evidence
        best_hyp = aic.index(min(aic))
        Delta_best = [a - aic[best_hyp] for a in aic]
        # we will only recommend H1 if it has the both the minimum AIC_c, and
        # the difference to H0 is larger than maxdel
        if (best_hyp == 1) and (Delta_best[0] > maxdel):
            return 1, Delta_best[0], (rss0/n_total, rss1/n_total)
        else:
            return 0, Delta_best[best_hyp], (rss0/n_total, rss1/n_total)

    @staticmethod
    def _search(data_and_params):
        """
        Parallelizable part of the search, search_network and search_catalog methods.
        """
        if len(data_and_params) == 5:
            x, y, kernel_size, kernel_size_min, maxdel = data_and_params
            check_only = None
        elif len(data_and_params) == 6:
            # added the optional fifth parameter of which indices of x to check
            x, y, kernel_size, kernel_size_min, maxdel, check_only = data_and_params
        else:
            raise RuntimeError("Passed invalid 'data_and_params' argument: "
                               f"{data_and_params}")
        # some checks
        assert isinstance(x, np.ndarray) and isinstance(y, np.ndarray), \
            f"'x' and 'y' need to be NumPy arrays, got {type(x)} and {type(y)}."
        # get sizes
        num_observations = x.shape[0]
        y = y.reshape(num_observations, 1 if y.ndim == 1 else -1)
        num_components = y.shape[1]
        # get valid array
        valid = np.isfinite(y)
        # make output arrays
        probs = np.empty((num_observations, num_components))
        probs[:] = np.NaN
        var0, var1 = probs.copy(), probs.copy()
        # loop over all columns
        for icomp in range(num_components):
            # loop through all rows, starting with a shrunken kernel at the edges
            # Beginning region
            halfwindow = 0
            for i in range(kernel_size // 2):
                if (check_only and (i not in check_only)) or \
                   (halfwindow*2 + 1 < kernel_size_min):
                    halfwindow += 1
                    continue
                hyp, Del, (rss0, rss1) = \
                    StepDetector.test_single(x[i-halfwindow:i+halfwindow+1],
                                             y[i-halfwindow:i+halfwindow+1, icomp],
                                             valid[i-halfwindow:i+halfwindow+1, icomp],
                                             maxdel=maxdel)
                if hyp == 1:
                    probs[i, icomp] = Del
                var0[i, icomp], var1[i, icomp] = rss0, rss1
                halfwindow += 1
            # Middle region
            assert halfwindow == kernel_size // 2
            range_main = range(halfwindow, num_observations - halfwindow)
            if check_only:
                range_main = [i for i in range_main if i in check_only]
            for i in range_main:
                hyp, Del, (rss0, rss1) = \
                    StepDetector.test_single(x[i-halfwindow:i+halfwindow+1],
                                             y[i-halfwindow:i+halfwindow+1, icomp],
                                             valid[i-halfwindow:i+halfwindow+1, icomp],
                                             maxdel=maxdel)
                if hyp == 1:
                    probs[i, icomp] = Del
                var0[i, icomp], var1[i, icomp] = rss0, rss1
            # Ending region
            for i in range(num_observations - halfwindow, num_observations):
                halfwindow -= 1
                if (check_only and (i not in check_only)) or \
                   (halfwindow*2 + 1 < kernel_size_min):
                    continue
                hyp, Del, (rss0, rss1) = \
                    StepDetector.test_single(x[i-halfwindow:i+halfwindow+1],
                                             y[i-halfwindow:i+halfwindow+1, icomp],
                                             valid[i-halfwindow:i+halfwindow+1, icomp],
                                             maxdel=maxdel)
                if hyp == 1:
                    probs[i, icomp] = Del
                var0[i, icomp], var1[i, icomp] = rss0, rss1
        # return
        if check_only:
            probs = probs[check_only, :].reshape(-1, num_components)
            var0 = var0[check_only, :].reshape(-1, num_components)
            var1 = var1[check_only, :].reshape(-1, num_components)
        return probs, var0, var1

    def search(self, x, y, maxdel=10):
        r"""
        Function that will search for steps in the data.
        Upon successful completion, it will return the relative step probabilities
        as well as the residuals variances of the two hypotheses tested
        (as reported by :meth:`~test_single`).

        Parameters
        ----------
        x : numpy.ndarray
            Input array of shape :math:`(\text{num_observations},)`.
            Should not contain NaNs.
        y : numpy.ndarray
            Input array of shape :math:`(\text{num_observations}, \text{num_components})`.
            Can contain NaNs.
        maxdel : float, optional
            Difference in AIC that should be considered not significantly better.
            (Refers to :math:`\Delta_i = \text{AIC}_{c,i} - \text{AIC}_{c,\text{min}}`.)

        Returns
        -------
        probabilities : numpy.ndarray
            Contains the relative probabilities array.
            Has shape :math:`(\text{num_observations}, \text{num_components})`.
        var0 : numpy.ndarray
            Contains the array of the residuals variance of the hypothesis
            that no step is present.
            Has shape :math:`(\text{num_observations}, \text{num_components})`.
        var1 : numpy.ndarray
            Contains the array of the residuals variance of the hypothesis
            that a step is present.
            Has shape :math:`(\text{num_observations}, \text{num_components})`.

        See Also
        --------
        :meth:`~test_single` : For more explanations about the return values.
        """
        # call individual search function and store result
        return StepDetector._search((x, y, self.kernel_size,
                                     self.kernel_size_min, maxdel))

    def search_network(self, net, ts_description, maxdel=10, threshold=20,
                       gap=2, gap_unit="D", no_pbar=False):
        r"""
        Function that searches for steps in an entire network (possibly in parallel),
        thresholds those probabilities, and identifies all the consecutive ranges in which
        steps happen over the network.

        Parameters
        ----------
        net : disstans.network.Network
            Network instance to operate on.
        ts_description : str
            :class:`~disstans.timeseries.Timeseries` description that will be analyzed.
        maxdel : float, optional
            Difference in AIC that should be considered not significantly better.
            (Refers to :math:`\Delta_i = \text{AIC}_{c,i} - \text{AIC}_{c,\text{min}}`.)
        threshold : float, optional
            Minimum :math:`\Delta_i \geq 0` that needs to be satisfied in order to be a step.
        gap : float, optional
            Maximum gap between identified steps to count as a continuous period
            of possible steps.
        gap_unit : str, optional
            Time unit of ``gap``.
        no_pbar : bool, optional
            Suppress the progress bar with ``True`` (default: ``False``).

        Returns
        -------
        step_table : pandas.DataFrame
            A DataFrame containing the columns ``'station'`` (its name), ``'time'``
            (a timestamp of the station) and ``'probability'`` (maximum :math:`\Delta_i`
            over all components for this timestamp), as well as ``var0``, ``var1``
            (the two hypotheses' residuals variances for the component of
            maximum step probability) and ``varred`` (the variance reduction in percent,
            ``(var0 - var1) / var0``).
        step_ranges : list
            A list of lists containing continuous periods over all stations of the potential
            steps as determined by ``gap`` and ``gap_unit``.
        """
        # get the stations who have this timeseries
        valid_stations = {name: station for name, station in net.stations.items()
                          if ts_description in station.timeseries}
        # make a list that will contain all individual result DataFrames
        step_tables = []
        # run parallelized StepDetector._search
        iterable_input = ((tvec_to_numpycol(station[ts_description].time),
                           station[ts_description].data.values,
                           self.kernel_size, self.kernel_size_min, maxdel)
                          for station in valid_stations.values())
        for name, station, (probs, var0, var1) in \
            zip(valid_stations.keys(), valid_stations.values(),
                tqdm(parallelize(StepDetector._search, iterable_input),
                     ascii=True, total=len(valid_stations), unit="station",
                     desc="Searching for steps", disable=no_pbar)):
            # find steps given the just calculated probabilities
            # setting the maximum number of steps to infinite to not miss anything
            steps = StepDetector.steps(probs, threshold, np.inf, False)
            # combine all data components and keep largest probability if the step
            # is present in multiple components
            unique_steps = np.sort(np.unique(np.concatenate(steps)))
            stepprobs = probs[unique_steps, :]
            stepsvar0, stepsvar1 = var0[unique_steps, :], var1[unique_steps, :]
            maxprobindices = np.expand_dims(np.argmax(stepprobs, axis=1), axis=1)
            maxstepprobs = np.take_along_axis(stepprobs, maxprobindices, axis=1).squeeze()
            maxstepvar0 = np.take_along_axis(stepsvar0, maxprobindices, axis=1).squeeze()
            maxstepvar1 = np.take_along_axis(stepsvar1, maxprobindices, axis=1).squeeze()
            # isolate the actual timestamps and add to the list of DataFrames
            steptimes = station[ts_description].time[unique_steps]
            step_tables.append(pd.DataFrame({"station": [name]*len(steptimes),
                                             "time": steptimes,
                                             "probability": maxstepprobs,
                                             "var0": maxstepvar0,
                                             "var1": maxstepvar1}))
            # this code could be used to create a model object and assign it to the station
            # mdl = disstans.models.Step(steptimes)
            # station.add_local_model(ts_description, "Detections", mdl)
        # combine individual DataFrames to one
        step_table = pd.concat(step_tables, ignore_index=True)
        # sort dataframe by probability
        step_table.sort_values(by="probability", ascending=False, inplace=True)
        # get coefficient of partial determination, i.e. how much the variance is reduced
        # (in percent) by including a step
        step_table["varred"] = (step_table["var0"] - step_table["var1"]) / step_table["var0"]
        # get the consecutive steptime ranges
        if not step_table.empty:
            unique_steps = np.sort(step_table["time"].unique())
            split = np.nonzero((np.diff(unique_steps) / Timedelta(1, gap_unit)) > gap)[0]
            split = np.concatenate([0, split + 1], axis=None)
            step_ranges = [unique_steps[split[i]:split[i + 1]] for i in range(split.size - 1)]
        else:
            step_ranges = []
        return step_table, step_ranges

    def search_catalog(self, net, ts_description, catalog, threshold=None,
                       gap=2, gap_unit="D", keep_nan_probs=True, no_pbar=False):
        r"""
        Search a dictionary of potential step times for each station in the dictionary
        and assess the probability for each one.

        Parameters
        ----------
        net : disstans.network.Network
            Network instance to operate on.
        ts_description : str
            :class:`~disstans.timeseries.Timeseries` description that will be analyzed.
        catalog : dict, pandas.DataFrame
            Dictionary where each key is a station name and its value is a list of
            :class:`~pandas.Timestamp` compatible potential times/dates.
            Alternatively, a DataFrame with at least the columns ``'station'`` and ``'time'``.
        threshold : float, optional
            Minimum :math:`\Delta_i \geq 0` that needs to be satisfied in order to be a step.
        gap : float, optional
            Maximum gap between identified steps to count as a continuous period
            of possible steps.
        gap_unit : str, optional
            Time unit of ``gap``.
        keep_nan_probs : bool, optional
            (Only applies to a DataFrame-type ``catalog`` input.)
            If a catalogued station is not in the network, or if a catalogued timestamp
            is after the available timeseries, no step probability can be calculated
            and the results will contain NaNs.
            If ``True`` (default), those entries will be kept in the output, and if
            ``False``, they will be dropped.
        no_pbar : bool, optional
            Suppress the progress bar with ``True`` (default: ``False``).

        Returns
        -------
        step_table : pandas.DataFrame
            A DataFrame containing the columns ``'station'`` (its name), ``'time'``
            (a timestamp of the station) and ``'probability'`` (maximum :math:`\Delta_i`
            over all components for this timestamp) for each potential step in ``catalog``,
            as well as ``var0`` and ``var1`` (the two hypotheses' residuals variances
            for the component of maximum step probability).
            If a DataFrame was passed as ``catalog``, a copy of that will be returned, with
            the added columns specified above.
        step_ranges : list
            A list of lists containing continuous periods over all stations of the potential
            steps as determined by ``gap`` and ``gap_unit``.
        """
        # hard-code maxdel to not filter out any item since we are asking about specific times
        maxdel = 0
        # get a simple dictionary representation if catalog was passed as dictionary
        # and set keep_nan_probs if not specified
        if isinstance(catalog, pd.DataFrame):
            assert all([col in catalog.columns for col in ["station", "time"]]), \
                "Invalid input 'catalog' DataFrame columns."
            catalog_df = catalog
            catalog = dict(catalog_df.groupby("station")["time"].apply(list))
            augment_df = True
        else:
            assert isinstance(catalog, dict), \
                "'catalog' must be either a dictionary or DataFrame."
            augment_df = False
        # for each station, find the first time index after a catalogued event
        # (alternatively, we could "add" a timestamp without an observation if
        # there isn't a timestamp already present - probably better, but harder)
        check_indices = {}
        # we also need to keep track of the originally requested time (for the output)
        catalog_timeexists = {sta_name: [False] * len(steptimes)
                              for sta_name, steptimes in catalog.items()}
        for sta_name, steptimes in catalog.items():
            # skip if station or timeseries not present
            if (sta_name not in net.stations.keys()) or \
               (ts_description not in net[sta_name].timeseries):
                continue
            check_indices[sta_name] = []
            for ist, st in enumerate(steptimes):
                index_after = (net[sta_name][ts_description].time >= st).tolist()
                try:
                    check_indices[sta_name].append(index_after.index(True))
                except ValueError:  # catalogued time is after the last timestamp
                    continue
                else:
                    catalog_timeexists[sta_name][ist] = True
        # make a list that will contain all individual result DataFrames
        step_tables = []
        # run parallelized StepDetector._search
        stations_overlap = list(check_indices.keys())
        iterable_input = ((tvec_to_numpycol(net[sta_name][ts_description].time),
                           net[sta_name][ts_description].data.values,
                           self.kernel_size, self.kernel_size_min,
                           maxdel, check_indices[sta_name])
                          for sta_name in stations_overlap)
        results_iterator = tqdm(parallelize(StepDetector._search, iterable_input),
                                ascii=True, total=len(stations_overlap), unit="station",
                                desc="Searching for steps", disable=no_pbar)
        for name, (probs, var0, var1) in zip(stations_overlap, results_iterator):
            # probs now contains a row for each catalog item
            # if the probability is NaN, AIC does not see evidence for a step,
            # if it is a float, then that's the likelihood of a step (always positive)
            # var0 and var1 can contain values regardless of the entry in probs
            has_steps = np.any(~np.isnan(probs), axis=1)
            maxstepprobs = probs[:, 0]
            maxstepvar0, maxstepvar1 = var0[:, 0], var1[:, 0]
            # maxstepprobs[has_steps] = np.nanmax(probs[has_steps, :], axis=1)
            maxprobindices = np.expand_dims(np.nanargmax(probs[has_steps, :], axis=1), axis=1)
            maxstepprobs[has_steps], maxstepvar0[has_steps], maxstepvar1[has_steps] = \
                np.take_along_axis(probs[has_steps, :], maxprobindices, axis=1).squeeze(), \
                np.take_along_axis(var0[has_steps, :], maxprobindices, axis=1).squeeze(), \
                np.take_along_axis(var1[has_steps, :], maxprobindices, axis=1).squeeze()
            # isolate the original timestamps and add to the list of DataFrames
            steptimes = [origtime for i, origtime in enumerate(catalog[name])
                         if catalog_timeexists[name][i]]
            step_tables.append(pd.DataFrame({"station": [name]*len(steptimes),
                                             "time": steptimes,
                                             "probability": maxstepprobs,
                                             "var0": maxstepvar0,
                                             "var1": maxstepvar1}))
        # combine individual DataFrames to one
        step_table = pd.concat(step_tables, ignore_index=True)
        # merge it with the input dataframe, if provided
        if augment_df:
            catalog_df["probability"] = np.NaN
            catalog_df["var0"] = np.NaN
            catalog_df["var1"] = np.NaN
            for _, row in step_table.iterrows():
                row_location = (catalog_df["station"] == row["station"]) & \
                               (catalog_df["time"] == row["time"])
                catalog_df.loc[row_location, ["probability", "var0", "var1"]] = \
                    row[["probability", "var0", "var1"]].values
            if not keep_nan_probs:
                catalog_df = catalog_df.dropna(how="all", subset=["probability", "var0", "var1"])
            step_table = catalog_df
        # sort
        step_table.sort_values(by="probability", ascending=False, inplace=True)
        # get coefficient of partial determination, i.e. how much the variance is reduced
        # (in percent) by including a step
        step_table["varred"] = (step_table["var0"] - step_table["var1"]) / step_table["var0"]
        # get the consecutive steptime ranges
        unique_steps = np.sort(step_table["time"].unique())
        split = np.nonzero((np.diff(unique_steps) / Timedelta(1, gap_unit)) > gap)[0]
        split = np.concatenate([0, split + 1], axis=None)
        step_ranges = [unique_steps[split[i]:split[i + 1]] for i in range(split.size - 1)]
        return step_table, step_ranges

    @staticmethod
    def steps(probabilities, threshold=2, maxsteps=np.inf, verbose=True):
        r"""
        Threshold the probabilities to return a list of steps.

        Parameters
        ----------
        threshold : float, optional
            Minimum :math:`\Delta_i \geq 0` that needs to be satisfied in order to be a step.
        maxsteps : int, optional
            Return at most ``maxsteps`` number of steps. Can be useful if a good value for
            ``threshold`` has not been found yet.
        verbose : bool, optional
            If ``True`` (default), print warnings when there will be a large number of
            steps identified given the ``threshold``.

        Returns
        -------
        steps : list
            Returns a list of :class:`~numpy.ndarray` arrays that contain the indices of steps
            for each component.
        """
        # initialize
        probabilities[np.isnan(probabilities)] = -1
        steps = []
        for icomp in range(probabilities.shape[1]):
            # find peaks above the threshold
            peaks, properties = find_peaks(probabilities[:, icomp], height=threshold)
            # if maxsteps is set, reduce steps to that number
            if peaks.size > maxsteps:
                largest_ix = np.argpartition(properties["peak_heights"], -maxsteps)[-maxsteps:]
                peaks = peaks[largest_ix]
                # warn about the new effective threshold
                if verbose:
                    print(f"In order to return at most {maxsteps} steps, the threshold has "
                          f"been increased to {properties['peak_heights'][largest_ix].min()}.")
            # warn if a large number of steps have been detected
            if verbose and (peaks.size / probabilities.shape[0] > 0.1):
                warnings.warn(f"In component {icomp}, using threshold={threshold} leads to "
                              f"{peaks.size / probabilities.shape[0]:.2%} of timestamps being "
                              "steps. Consider setting a higher threshold.", stacklevel=2)
            steps.append(peaks)
        return steps



"""
This module contains the :class:`~Timeseries` base class and other
formats included by default in DISSTANS.
"""

import numpy as np
import pandas as pd
from warnings import warn
from copy import deepcopy

from .tools import get_cov_dims, make_cov_index_map, get_cov_indices


class Timeseries():
    """
    Object that expands the functionality of a :class:`~pandas.DataFrame` object
    for better integration into DISSTANS. Apart from the data itself, it contains
    information about the source and units of the data. It also performs input
    checks and uses property setters/getters to ensure consistency.

    Also enables the ability to perform math on timeseries directly.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The timeseries' data as a DataFrame. The index should be time, whereas
        data columns can be both data and their uncertainties.
    src : str
        Source description.
    data_unit : str
        Data unit.
    data_cols : list
        List of strings with the names of the columns of ``dataframe`` that
        contain the data.
        The length cooresponds to the number of components :attr:`~num_components`.
    var_cols : list, optional
        List of strings with the names of the columns of ``dataframe`` that contain the
        data's variance.
        Must have the same length as ``data_cols``.
        Defaults to no data variance columns.
    cov_cols : list, optional
        List of strings with the names of the columns of ``dataframe`` that contain
        the data's covariance.
        Must have length ``(num_components * (num_components - 1)) / 2``, where
        the order of the elements is determined by their row-by-row, sequential
        position in the covariance matrix (see Notes).
        Defaults to no covariance columns.

    Notes
    -----

    In terms of mapping the covariance matrix of observations into the format
    for the :class:`~Timeseries` class, consider this example for observations with
    three components:

    +-------------------+-------------------+-------------------+
    | ``var_cols[0]``   | ``cov_cols[0]``   | ``cov_cols[1]``   |
    +-------------------+-------------------+-------------------+
    | (symmetric)       | ``var_cols[1]``   | ``cov_cols[2]``   |
    +-------------------+-------------------+-------------------+
    | (symmetric)       | (symmetric)       | ``var_cols[2]``   |
    +-------------------+-------------------+-------------------+
    """
    def __init__(self, dataframe, src, data_unit, data_cols,
                 var_cols=None, cov_cols=None):
        # input type checks
        assert isinstance(dataframe, pd.DataFrame)
        assert isinstance(src, str)
        assert isinstance(data_unit, str)
        assert isinstance(data_cols, list) and all([isinstance(dcol, str) for dcol in data_cols])
        assert all([dcol in dataframe.columns for dcol in data_cols])
        # save to private instance variables
        self._df = dataframe
        self._src = src
        self._data_unit = data_unit
        self._data_cols = data_cols
        # additional checks for var_cols
        if var_cols is not None:
            assert isinstance(var_cols, list) and \
                all([isinstance(vcol, str) for vcol in var_cols]), \
                f"If not None, 'var_cols' must be a list of strings, got {var_cols}."
            assert all([vcol in dataframe.columns for vcol in var_cols]), \
                "All entries in 'var_cols' must be present in the dataframe."
            assert len(self._data_cols) == len(var_cols), \
                "If passing variance columns, " \
                "the list needs to have the same length as the data columns one."
        self._var_cols = var_cols
        # additional checks for cov_cols
        if cov_cols is not None:
            assert self.num_components > 1, \
                "In order to set covariances, the timeseries needs at least 2 components."
            assert cov_cols is not None, \
                "If setting covariances, the standard deviations must also be set."
            assert isinstance(cov_cols, list) and \
                all([isinstance(ccol, str) for ccol in cov_cols]), \
                f"If not None, 'cov_cols' must be a list of strings, got {cov_cols}."
            assert all([ccol in dataframe.columns for ccol in cov_cols]), \
                "All entries in 'cov_cols' must be present in the dataframe."
            cov_dims = get_cov_dims(self.num_components)
            assert len(cov_cols) == cov_dims, \
                "If passing covariance columns, the list needs to have the appropriate " \
                "length given the data columns length. " \
                f"Expected {cov_dims}, got {len(cov_cols)}."
            # everything should be alright, make index maps
            index_map, var_cov_map = make_cov_index_map(self.num_components)
            self.index_map = index_map
            """
            Matrix that contains the rolling indices of each matrix element used by
            :func:`~disstans.tools.get_cov_indices`.
            """
            self.var_cov_map = var_cov_map
            """
            Contains the column indices needed to create the full variance-covariance
            matrix for a single time.
            """
        self._cov_cols = cov_cols

    def __str__(self):
        """
        Special function that returns a readable summary of the timeseries.
        Accessed, for example, by Python's ``print()`` built-in function.

        Returns
        -------
        info : str
            Timeseries summary.
        """
        info = f"Timeseries\n - Source: {self.src}\n - Units: {self.data_unit}" \
               f"\n - Shape: {self.shape}\n - Data: {[key for key in self.data_cols]}"
        if self.var_cols is not None:
            info += f"\n - Variances: {[key for key in self.var_cols]}"
        if self.cov_cols is not None:
            info += f"\n - Covariances: {[key for key in self.cov_cols]}"
        return info

    def __getitem__(self, columns):
        """
        Convenience special function that provides a shorthand notation
        to access the timeseries' columns.

        Parameters
        ----------
        columns : str, list
            String or list of strings of the columns to return.

        Returns
        -------
        pandas.Series, pandas.DataFrame
            Returns the requested data as a Series (if a single column) or
            DataFrame (if multiple columns).

        Example
        -------
        If ``ts`` is a :class:`~Timeseries` instance and ``columns`` a list
        of column names, the following two are equivalent::

            ts.df[columns]
            ts[ts_description]
        """
        if not isinstance(columns, str) and \
           (not isinstance(columns, list) or
           not all([isinstance(col, str) for col in columns])):
            raise KeyError("Error when accessing data in timeseries: 'column' must be a "
                           f"string or list of strings, given was {columns}.")
        return self.df[columns]

    @property
    def src(self):
        """ Source information. """
        return self._src

    @src.setter
    def src(self, new_src):
        if not isinstance(new_src, str):
            raise TypeError(f"New 'src' attribute has to be a string, got {type(new_src)}.")
        self._src = new_src

    @property
    def data_unit(self):
        """ Data unit. """
        return self._data_unit

    @data_unit.setter
    def data_unit(self, new_data_unit):
        if not isinstance(new_data_unit, str):
            raise TypeError("New 'data_unit' attribute has to be a string, "
                            f"got {type(new_data_unit)}.")
        self._data_unit = new_data_unit

    @property
    def data_cols(self):
        """ List of the column names in :attr:`~df` that contain data. """
        return self._data_cols

    @data_cols.setter
    def data_cols(self, new_data_cols):
        assert isinstance(new_data_cols, list) and \
            all([isinstance(dcol, str) for dcol in new_data_cols]), \
            "New 'data_cols' attribute must be a list of strings of the same length as the " \
            f"current 'data_cols' ({len(self._data_cols)}), got {new_data_cols}."
        self._df.rename(columns={old_col: new_col for old_col, new_col
                                 in zip(self._data_cols, new_data_cols)},
                        errors='raise', inplace=True)
        self._data_cols = new_data_cols

    @property
    def var_cols(self):
        """ List of the column names in :attr:`~df` that contain data variance. """
        return self._var_cols

    @var_cols.setter
    def var_cols(self, new_var_cols):
        assert self._var_cols is not None, \
            "No variance columns found that can be renamed."
        assert isinstance(new_var_cols, list) and \
            all([(vcol is None) or isinstance(vcol, str) for vcol in new_var_cols]), \
            "New 'var_cols' attribute must be a list of strings or Nones of the same length " \
            f"as the current 'var_cols' ({len(self._var_cols)}), got {new_var_cols}."
        for ivcol, vcol in enumerate(new_var_cols):
            if vcol is None:
                new_var_cols[ivcol] = self._var_cols[ivcol]
        self._df.rename(columns={old_col: new_col for old_col, new_col
                                 in zip(self._var_cols, new_var_cols)},
                        errors='raise')
        self._var_cols = new_var_cols

    @property
    def cov_cols(self):
        """ List of the column names in :attr:`~df` that contain data covariances. """
        return self._cov_cols

    @cov_cols.setter
    def cov_cols(self, new_cov_cols):
        assert self._cov_cols is not None, \
            "No covariance columns found that can be renamed."
        assert isinstance(new_cov_cols, list) and \
            all([(ccol is None) or isinstance(ccol, str) for ccol in new_cov_cols]), \
            "New 'cov_cols' attribute must be a list of strings or Nones of the same length " \
            f"as the current 'cov_cols' ({len(self._cov_cols)}), got {new_cov_cols}."
        for iccol, ccol in enumerate(new_cov_cols):
            if ccol is None:
                new_cov_cols[iccol] = self._cov_cols[iccol]
        self._df.rename(columns={old_col: new_col for old_col, new_col
                                 in zip(self._cov_cols, new_cov_cols)},
                        errors='raise')
        self._cov_cols = new_cov_cols

    @property
    def num_components(self):
        """ Number of data columns. """
        return len(self._data_cols)

    @property
    def df(self):
        """ The entire timeseries' DataFrame. """
        return self._df

    @property
    def num_observations(self):
        """ Number of observations (rows in :attr:`~df`). """
        return self._df.shape[0]

    @property
    def data(self):
        """ View of only the data columns in :attr:`~df`. """
        return self._df.loc[:, self._data_cols]

    @data.setter
    def data(self, new_data):
        self._df.loc[:, self._data_cols] = new_data

    @property
    def sigmas(self):
        """ View of only the data standard deviation columns in :attr:`~df`. """
        return self.vars ** (1/2)

    @sigmas.setter
    def sigmas(self, new_sigma):
        self.vars = new_sigma ** 2

    @property
    def vars(self):
        """ Returns the variances from :attr:`~df`. """
        if self._var_cols is None:
            raise ValueError("No variance columns present to return.")
        return self._df.loc[:, self._var_cols]

    @vars.setter
    def vars(self, new_var):
        if self._var_cols is None:
            raise ValueError("No variance columns present to set.")
        self.df.loc[:, self._var_cols] = new_var

    @property
    def covs(self):
        """ Returns the covariances from :attr:`~df`. """
        if self._cov_cols is None:
            raise ValueError("No covariance columns present to return.")
        return self._df.loc[:, self._cov_cols]

    @covs.setter
    def covs(self, new_cov):
        if self._cov_cols is None:
            raise ValueError("No covariance columns present to set.")
        self.df.loc[:, self._cov_cols] = new_cov

    @property
    def var_cov(self):
        """
        Returns the variance as well as covariance columns from :attr:`~df`, to be indexed
        by :attr:`~var_cov_map` to yield the full variance-covariance matrix.
        """
        if self._var_cols is None:
            raise ValueError("No variance columns present to return.")
        if self._cov_cols is None:
            raise ValueError("No covariance columns present to return.")
        return self._df.loc[:, self._var_cols + self._cov_cols]

    @var_cov.setter
    def var_cov(self, new_var_cov):
        cov_dims = get_cov_dims(self.num_components)
        assert new_var_cov.shape[1] == self.num_components + cov_dims, \
            "Setting 'var_cov' requires a column for each variance (first half of array, " \
            f"{self.num_components}) and covariance (second half, {cov_dims})."
        self.vars = new_var_cov[:, :self.num_components]
        self.covs = new_var_cov[:, self.num_components:]

    @property
    def time(self):
        """ Timestamps of the timeseries (index of :attr:`~df`). """
        return self._df.index

    @property
    def shape(self):
        r"""
        Returns the shape tuple (similar to NumPy) of the timeseries, which is of shape
        :math:`(\text{n_observations},\text{n_components})`.
        """
        return (self.num_observations, self.num_components)

    @property
    def length(self):
        """
        Returns the length of the timeseries.
        """
        return self.time.max() - self.time.min()

    @property
    def reliability(self):
        """
        Returns the reliability (between 0 and 1) defined as the number of available
        observations divided by the the number of expected observations. The expected
        observations are calculated by taking the median timespan between observations,
        and then dividing the total time span by that timespan.

        (Essentially, this assumes that there are not any "close-by" observation, e.g.
        two observation for the same day but a different hour in a dataset of otherwise
        daily observations.)
        """
        # get median time difference
        med_tdiff = np.median(np.diff(self.time.values))
        # expected observations
        exp_obs = int(np.round(self.length / med_tdiff + 1))
        # return ratio
        return self.num_observations / exp_obs

    def cov_at(self, t):
        """
        Returns the covariance matrix of the timeseries at a given time or index.

        Parameters
        ----------
        t : pandas.Timestamp, str, int
            A timestamp or timestamp-convertable string to return the covariance
            matrix for. Alternatively, an integer index.

        Returns
        -------
        cov_mat : numpy.ndarray
            The full covariance matrix at time ``t``.
        """
        # input checks
        assert self._var_cols is not None, \
            "No variances available."
        if isinstance(t, str):
            t = pd.Timestamp(t)
        elif isinstance(t, int):
            t = self.time[t]
        assert self.time.min() <= t <= self.time.max(), \
            f"Time {t} not inside available timespan."
        assert np.any(np.isfinite(self.vars[t].values)), \
            f"No variances set at time {t}."
        # only variance is set
        if self._cov_cols is None:
            cov_mat = np.diag(self.vars[t].values)
        # full covariance matrix is available
        else:
            cov_mat = np.reshape(self.var_cov[t].values[0, self.var_cov_map],
                                 (self.num_components, self.num_components))
        return cov_mat

    def cut(self, t_min=None, t_max=None, i_min=None, i_max=None, keep_inside=True):
        """
        Cut the timeseries to contain only data between certain times or indices.
        If both a minimum (maximum) timestamp or index is provided, the later (earlier,
        respectively) one is used (i.e., the more restrictive one).
        Also provides the reverse operation, i.e. only removing data between dates.

        This operation changes the timeseries in-place; if it should be done on a
        new timeseries, use :meth:`~copy` first.

        Parameters
        ----------
        t_min : pandas.Timestamp, str, optional
            A timestamp or timestamp-convertable string of the earliest observation
            to keep.
        t_max : pandas.Timestamp, str, optional
            A timestamp or timestamp-convertable string of the latest observation
            to keep.
        i_min : int, optional
            The index of the earliest observation to keep.
        i_max : int, optional
            The index of the latest observation to keep.
        keep_inside : bool, optional
            If ``True`` (default), keeps data inside of the specified date range.
            If ``False``, keeps only data outside the specified date range.
        """
        assert any([t_min, t_max, i_min, i_max]), "No cutting operation defined."
        # convert to timestamps (if not already)
        if i_min:
            t_from_i_min = self.time[i_min]
        if i_max:
            t_from_i_max = self.time[i_max]
        if t_min:
            t_min = pd.Timestamp(t_min)
        if t_max:
            t_max = pd.Timestamp(t_max)
        # get more restrictive cutoffs
        if i_min and t_min:
            cut_min = max([t_from_i_min, t_min])
        elif i_min or t_min:
            cut_min = t_min if t_min else t_from_i_min
        else:
            cut_min = self.time[0]
        if i_max and t_max:
            cut_max = min([t_from_i_max, t_max])
        elif i_max or t_max:
            cut_max = t_max if t_max else t_from_i_max
        else:
            cut_max = self.time[-1]
        # cut the dataframe directly
        if keep_inside:
            self._df = self._df[(self.time >= cut_min) & (self.time <= cut_max)]
        else:
            self._df = self._df[(self.time < cut_min) | (self.time > cut_max)]

    def add_uncertainties(self, timeseries=None,
                          var_data=None, var_cols=None, cov_data=None, cov_cols=None):
        """
        Add variance and covariance data and column names to the timeseries.

        Parameters
        ----------
        timeseries : disstans.timeseries.Timeseries
            Another timeseries object that contains uncertainty information.
            If set, the function will ignore the rest of the arguments.
        var_data : numpy.ndarray, optional
            New data variance.
        var_cols : list, optional
            List of variance column names.
        cov_data : numpy.ndarray, optional
            New data covariance.
            Setting this but not ``var_data`` requires there to already be data variance.
        cov_cols : list, optional
            List of covariance column names.

        Notes
        -----
        If ``ts`` is a :class:`~disstans.timeseries.Timeseries` instance, just using::

            ts.vars = new_variance
            ts.covs = new_covariance

        will only work when the respective columns already exist in the dataframe.
        (This is the same behavior for renaming variance columns that do not exist.)
        If they do not exist, the calls will results in an error because no column names exist,
        in an effort to make the inner workings more transparent and rigorous.

        This function allows to override the default behavior, and can also generate column
        names by itself if none are specified.
        """
        # check if input is a Timeseries, override rest of input arguments
        if timeseries is not None:
            if not isinstance(timeseries, Timeseries):
                raise TypeError("Cannot use uncertainty data: "
                                "'timeseries' is not a Timeseries object.")
            # get variance data and description
            if timeseries.var_cols is not None:
                var_data = timeseries.vars[timeseries.time.isin(self.time)].values
                var_cols = timeseries.var_cols
            else:
                var_data = None
                var_cols = None
            # get covariance data and description
            if timeseries.cov_cols is not None:
                cov_data = timeseries.covs[timeseries.time.isin(self.time)].values
                cov_cols = timeseries.cov_cols
            else:
                cov_data = None
                cov_cols = None
        # add variance
        if var_data is not None:
            assert isinstance(var_data, np.ndarray) and var_data.shape == self.shape, \
                   "'var_data' must be a NumPy array of the same shape as the data " \
                   f"{self.shape}, got " \
                   f"{var_data.shape if isinstance(var_data, np.ndarray) else var_data}."
            if var_cols is None:
                var_cols = [dcol + "_sigma" for dcol in self.data_cols]
            else:
                assert (isinstance(var_cols, list) and len(var_cols) == self.num_components and
                        all([isinstance(vcol, str) for vcol in var_cols])), \
                       "New 'var_cols' attribute must be a list of strings of the same length " \
                       f"as the current 'data_cols' ({len(self._data_cols)}), got {var_cols}."
            for ivcol, vcol in enumerate(var_cols):
                self._df[vcol] = var_data[:, ivcol]
            self._var_cols = var_cols
        # add covariance, check for variance
        if cov_data is not None:
            assert self._var_cols is not None, \
                "Cannot set covariance data without first adding variance data."
            cov_dims = get_cov_dims(self.num_components)
            assert (isinstance(cov_data, np.ndarray)
                    and cov_data.shape[0] == self.num_observations
                    and cov_data.shape[1] == cov_dims), \
                "'cov_data' must be a NumPy array of the same length as the data " \
                "and width corresponding to all possible covariances. Expected shape " \
                f"({self.num_observations}, {cov_dims}), got " \
                f"{cov_data.shape if isinstance(cov_data, np.ndarray) else cov_data}."
            if cov_cols is None:
                cov_cols = []
                for i1 in range(self.num_components):
                    for i2 in range(i1 + 1, self.num_components):
                        cov_cols.append(f"{self.data_cols[i1]}_{self.data_cols[i2]}_cov")
            else:
                assert (isinstance(cov_cols, list) and len(cov_cols) == cov_dims and
                        all([isinstance(vcol, str) for vcol in cov_cols])), \
                    "New 'cov_cols' attribute must be a list of strings of length " \
                    "corresponding to all possible covariances " \
                    f"( expected {cov_dims}, got {len(cov_cols)}."
            for iccol, ccol in enumerate(cov_cols):
                self._df[ccol] = cov_data[:, iccol]
            self.index_map, self.var_cov_map = make_cov_index_map(self.num_components)
            self._cov_cols = cov_cols

    def copy(self, only_data=False, src=None):
        """
        Return a deep copy of the timeseries instance.

        Parameters
        ----------
        only_data : bool, optional
            If ``True``, only copy the data columns and ignore any uncertainty information.
            Defaults to ``False``.
        src : str, optional
            Set a new source information attribute for the copy.
            Uses the current one by default.

        Returns
        -------
        Timeseries
            The copy of the timeseries instance.
        """
        new_name = deepcopy(self._src) if src is None else src
        if not only_data:
            return Timeseries(self._df.copy(), new_name, deepcopy(self._data_unit),
                              deepcopy(self._data_cols), deepcopy(self._var_cols),
                              deepcopy(self._cov_cols))
        else:
            return Timeseries(self._df[self._data_cols].copy(), new_name,
                              deepcopy(self._data_unit), deepcopy(self._data_cols), None, None)

    def convert_units(self, factor, new_data_unit):
        """
        Convert the data and covariances to a new data unit by providing a
        conversion factor.

        Parameters
        ----------
        factor : float
            Factor to multiply the data by to obtain the data in the new units.
        new_data_unit : str
            New data unit to be saved in the :attr:`~data_cols` attribute.
        """
        # input checks
        try:
            factor = float(factor)
        except TypeError as e:
            raise TypeError(f"'factor' and has to be a scalar, got {type(factor)}."
                            ).with_traceback(e.__traceback__) from e
        # update data unit (contains input check)
        self.data_unit = new_data_unit
        # convert data
        self._df.loc[:, self._data_cols] *= factor
        # convert variances
        if self._var_cols is not None:
            self._df.loc[:, self._var_cols] *= factor**2
        # convert covariances
        if self._cov_cols is not None:
            self._df.loc[:, self._cov_cols] *= factor**2

    def mask_out(self, dcol):
        """
        Mask out an entire data column (and if present, its uncertainty column) by setting
        the entire column to ``NaN``. Converts it to a sparse representation to save memory.

        Parameters
        ----------
        dcol : str
            Name of the data column to mask out.
        """
        icomp = self._data_cols.index(dcol)
        self._df[dcol] = np.NaN
        self._df[dcol] = self._df[dcol].astype(pd.SparseDtype(dtype=float))
        if self._var_cols is not None:
            vcol = self._var_cols[icomp]
            self._df[vcol] = np.NaN
            self._df[vcol] = self._df[vcol].astype(pd.SparseDtype(dtype=float))
            if self._cov_cols is not None:
                for iccol in get_cov_indices(icomp, index_map=self.index_map):
                    ccol = self._cov_cols[iccol]
                    self._df[ccol] = np.NaN
                    self._df[ccol] = self._df[ccol].astype(pd.SparseDtype(dtype=float))

    @staticmethod
    def prepare_math(left, right, operation):
        r"""
        Tests two timeseries' ability to be cast together in a mathematical operation,
        and returns output characteristics.
        Currently, only addition, subtraction, multiplication, and division are supported.

        All uncertainty information is lost during mathematical operations.

        One of the objects can be a NumPy array. In this case, the array has to have
        the exact same shape as the data in the Timeseries instance.
        Furthermore, the resulting Timeseries object will have the same :attr:`~src`,
        :attr:`~data_unit` and :attr:`~data_cols` attributes (instead of a combination
        of both).

        Warning
        -------
        This method is called under-the-hood whenever a mathematical operation is
        performed, and should not need to be used by normal users.

        Parameters
        ----------
        left : Timeseries or numpy.ndarray
            Left term of the operation.
        right : Timeseries or numpy.ndarray
            Right term of the operation.
        operation : str
            Operation to perform.
            Possible values are ``'+'``, ``'-'``, ``'*'`` and ``'/'``.

        Returns
        -------
        left_data : numpy.ndarray
            View of the 2D left data array of the operation
            with shape (``len(out_time)``, :attr:`~num_components`).
        right_data : numpy.ndarray
            View of the 2D right data array of the operation
            with shape (``len(out_time)``, :attr:`~num_components`).
        out_src : str
            Combines the sources of each object to a new string.
        out_data_unit : str
            Combines the data units of each object into a new unit.
        out_data_cols : list
            List of strings containing the new data column names.
        out_time : pandas.Index
            Index object containing the indices of all timestamps common to both.

        Raises
        ------
        TypeError
            If one of the operands is not a :class:`~Timeseries` or :class:`~numpy.ndarray`,
            or if both are :class:`~numpy.ndarray` (since then this function would never be
            called anyway).
        ValueError
            If the number of data columns is not equal between the two operands,
            or if the data units are not the same adding or subtracting.
        AssertionError
            If one of the operands is a NumPy array but does not have the same number of
            rows as the other operand.

        See Also
        --------
        __add__ : Addition for two Timeseries or a Timeseries and a NumPy array
        __radd__ : Addition for a NumPy array and a Timeseries.
        __sub__ : Subtraction for two Timeseries or a Timeseries and a NumPy array
        __rsub__ : Subtraction for a NumPy array and a Timeseries.
        __mul__ : Multiplication for two Timeseries or a Timeseries and a NumPy array
        __rmul__ : Multiplication for a NumPy array and a Timeseries.
        __truediv__ : Division for two Timeseries or a Timeseries and a NumPy array
        __rtruediv__ : Division for a NumPy array and a Timeseries.
        """
        # check for valid operator
        if operation not in ["+", "-", "*", "/"]:
            raise NotImplementedError("Timeseries math problem: "
                                      f"unknown operation '{operation}'.")
        # check for compatible types
        for operand, position in zip([left, right], ["Left", "Right"]):
            if (not isinstance(operand, Timeseries)) and (not isinstance(operand, np.ndarray)):
                raise TypeError(f"{position} operand has to be either a '{Timeseries}' or "
                                f"'{np.ndarray}' object, got {type(operand)}")
        if all([isinstance(operand, np.ndarray) for operand in [left, right]]):
            raise TypeError(f"At least one of the operands has to be a '{Timeseries}' object, "
                            f"got two '{np.ndarray}' objects.")
        # check for same dimensions
        if isinstance(left, Timeseries):
            len_left_data_cols = left.num_components
        else:
            len_left_data_cols = left.shape[1]
        if isinstance(right, Timeseries):
            len_right_data_cols = right.num_components
        else:
            len_right_data_cols = right.shape[1]
        if len_left_data_cols != len_right_data_cols:
            raise ValueError("Timeseries math problem: conflicting number of data columns "
                             f"({len_left_data_cols} and {len_right_data_cols}).")
        # get output attributes
        if isinstance(left, np.ndarray):
            assert left.shape[0] == right.num_observations, \
                "Timeseries math problem: conflicting number of observation " + \
                f"({left.shape[0]} and {right.num_observations})."
            out_time = right.time
            out_src = right.src
            out_data_unit = right.data_unit
            out_data_cols = right.data_cols
            left_data = left
            right_data = right.data.loc[out_time, :].values
        elif isinstance(right, np.ndarray):
            assert left.num_observations == right.shape[0], \
                "Timeseries math problem: conflicting number of observation " + \
                f"({left.num_observations} and {right.shape[0]})."
            out_time = left.time
            out_src = left.src
            out_data_unit = left.data_unit
            out_data_cols = left.data_cols
            left_data = left.data.loc[out_time, :].values
            right_data = right
        else:
            # time index is the intersection
            out_time = left.time.intersection(right.time, sort=None)
            # check compatible units (+-) or define new one (*/)
            # define new src and data_cols
            if operation in ["+", "-"]:
                if left.data_unit != right.data_unit:
                    raise ValueError("Timeseries math problem: conflicting data units "
                                     f"'{left.data_unit}' and '{right.data_unit}'.")
                else:
                    out_data_unit = left.data_unit
                    out_src = f"{left.src}{operation}{right.src}"
                    out_data_cols = [f"{lcol}{operation}{rcol}" for lcol, rcol
                                     in zip(left.data_cols, right.data_cols)]
            elif operation in ["*", "/"]:
                out_data_unit = f"({left.data_unit}){operation}({right.data_unit})"
                out_src = f"({left.src}){operation}({right.src})"
                out_data_cols = [f"({lcol}){operation}({rcol})" for lcol, rcol
                                 in zip(left.data_cols, right.data_cols)]
            left_data = left.data.loc[out_time, :].values
            right_data = right.data.loc[out_time, :].values
        # return data unit and column names
        return left_data, right_data, out_src, out_data_unit, out_data_cols, out_time

    def __add__(self, other):
        """
        Special function that allows two timeseries instances (or a timeseries and
        an equivalently shaped NumPy array) to be added together element-wise.

        Parameters
        ----------
        other : Timeseries
            Timeseries to add to instance.

        Returns
        -------
        Timeseries
            New timeseries object containing the sum of the two timeseries.

        See Also
        --------
        prepare_math
            Prepares the two instances for the mathematical operation. Refer to it
            for more details about how the two objects are cast together.

        Example
        -------
        Add two :class:`~Timeseries` ``ts1`` and ``ts2`` and save the result as ``ts3``::

            ts3 = ts1 + ts2
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(self, other, '+')
        return Timeseries(pd.DataFrame(left_data + right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __radd__(self, other):
        """
        Reflected operation of :meth:`~__add__` (necessary if first operand is a NumPy array).
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(other, self, '+')
        return Timeseries(pd.DataFrame(left_data + right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __sub__(self, other):
        """
        Special function that allows a timeseries instance (or a timeseries and
        an equivalently shaped NumPy array) to be subtracted from another element-wise.

        Parameters
        ----------
        other : Timeseries
            Timeseries to subtract from instance.

        Returns
        -------
        Timeseries
            New timeseries object containing the difference of the two timeseries.

        See Also
        --------
        prepare_math
            Prepares the two instances for the mathematical operation. Refer to it
            for more details about how the two objects are cast together.

        Example
        -------
        Subtract two :class:`~Timeseries` ``ts1`` and ``ts2`` and save the result as ``ts3``::

            ts3 = ts1 - ts2
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(self, other, '-')
        return Timeseries(pd.DataFrame(left_data - right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __rsub__(self, other):
        """
        Reflected operation of :meth:`~__sub__` (necessary if first operand is a NumPy array).
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(other, self, '-')
        return Timeseries(pd.DataFrame(left_data - right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __mul__(self, other):
        """
        Special function that allows two timeseries instances (or a timeseries and
        an equivalently shaped NumPy array) to be multiplied together element-wise.

        Parameters
        ----------
        other : Timeseries
            Timeseries to multiply to instance.

        Returns
        -------
        Timeseries
            New timeseries object containing the product of the two timeseries.

        See Also
        --------
        prepare_math
            Prepares the two instances for the mathematical operation. Refer to it
            for more details about how the two objects are cast together.

        Example
        -------
        Multiply two :class:`~Timeseries` ``ts1`` and ``ts2`` and save the result as ``ts3``::

            ts3 = ts1 * ts2
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(self, other, '*')
        return Timeseries(pd.DataFrame(left_data * right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __rmul__(self, other):
        """
        Reflected operation of :meth:`~__mul__` (necessary if first operand is a NumPy array).
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(other, self, '*')
        return Timeseries(pd.DataFrame(left_data * right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __truediv__(self, other):
        """
        Special function that allows a timeseries instances (or a timeseries and
        an equivalently shaped NumPy array) to be divided by another element-wise.

        Parameters
        ----------
        other : Timeseries
            Timeseries to divide instance by.

        Returns
        -------
        Timeseries
            New timeseries object containing the quotient of the two timeseries.

        See Also
        --------
        prepare_math
            Prepares the two instances for the mathematical operation. Refer to it
            for more details about how the two objects are cast together.

        Example
        -------
        Divide two :class:`~Timeseries` ``ts1`` and ``ts2`` and save the result as ``ts3``::

            ts3 = ts1 / ts2
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(self, other, '/')
        return Timeseries(pd.DataFrame(left_data / right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def __rtruediv__(self, other):
        """
        Reflected operation of :meth:`~__truediv__` (necessary if first operand is a NumPy array).
        """
        left_data, right_data, out_src, out_unit, out_data_cols, out_time = \
            Timeseries.prepare_math(other, self, '/')
        return Timeseries(pd.DataFrame(left_data / right_data,
                                       index=out_time, columns=out_data_cols),
                          out_src, out_unit, out_data_cols)

    def get_arch(self):
        """
        Build a dictionary describing the architecture of this timeseries,
        to be used when creating a network JSON configuration file.

        Without subclassing :class:`~Timeseries`, this function will return an
        empty dictionary by default, since it is unknown how to recreate
        a general Timeseries object from just a JSON-compatible dictionary.

        See Also
        --------
        disstans.network.Network.to_json : Export the Network configuration as a JSON file.
        disstans.timeseries.Timeseries.get_arch
            Get the architecture dictionary of a :class:`~disstans.timeseries.Timeseries` instance.
        disstans.models.Model.get_arch
            Get the architecture dictionary of a :class:`~disstans.models.Model` instance.
        """
        return {}

    @classmethod
    def from_fit(cls, data_unit, data_cols, fit):
        """
        Import a fit dictionary and create a Timeseries instance.

        Parameters
        ----------
        data_unit : str
            Data unit.
        data_cols : list
            List of strings containing the data column names.
            Uncertainty column names are generated by adding a '_var'.
        fit : dict
            Dictionary with the keys ``'time'``, ``'fit'``, ``'var'`` and ``'cov'``
            (the latter two can be set to ``None``).

        Returns
        -------
        Timeseries
            Timeseries instance created from ``fit``.

        See Also
        --------
        disstans.models.Model.evaluate
            Evaluating a model produces the fit dictionary.
        """
        df_data = {dcol: fit["fit"][:, icol] for icol, dcol in enumerate(data_cols)}
        var_cols = None
        cov_cols = None
        if fit["var"] is not None:
            var_cols = [dcol + "_var" for dcol in data_cols]
            df_data.update({vcol: fit["var"][:, icol] for icol, vcol in enumerate(var_cols)})
            if fit["cov"] is not None:
                cov_cols = []
                num_components = len(data_cols)
                for i1 in range(num_components):
                    for i2 in range(i1 + 1, num_components):
                        cov_cols.append(f"{data_cols[i1]}_{data_cols[i2]}_cov")
                df_data.update({ccol: fit["cov"][:, icol] for icol, ccol in enumerate(cov_cols)})
        df = pd.DataFrame(data=df_data, index=fit["time"])
        return cls(df, "fitted", data_unit, data_cols, var_cols, cov_cols)

    @classmethod
    def from_array(cls, timevector, data, src, data_unit, data_cols,
                   var=None, var_cols=None, cov=None, cov_cols=None):
        r"""
        Constructor method to create a :class:`~Timeseries` instance from a NumPy
        :class:`~numpy.ndarray`.

        Parameters
        ----------
        timevector : pandas.Series, pandas.DatetimeIndex
            :class:`~pandas.Series` of :class:`~pandas.Timestamp` or alternatively a
            :class:`~pandas.DatetimeIndex` containing the timestamps of each observation.
        data : numpy.ndarray
            2D NumPy array of shape :math:`(\text{n_observations},\text{n_components})`
            containing the data.
        src : str
            Source description.
        data_unit : str
            Data unit.
        data_cols : list
            List of strings with the names of the columns of ``data``.
        var : numpy.ndarray, optional
            2D NumPy array of shape :math:`(\text{n_observations},\text{n_components})`
            containing the data variances.
            Defaults to no data uncertainty.
        var_cols : list, optional
            List of strings with the names of the columns of ``data`` that contain the
            data's variance.
            Must have the same length as ``data_cols``.
            If ``var`` is given but ``var_cols`` is not, it defaults to appending
            ``'_var'`` to ``data_cols``.
        cov : numpy.ndarray, optional
            2D NumPy array of shape :math:`(\text{n_observations},\text{n_components})`
            containing the data covariances (as defined in :class:`~Timeseries`).
            Defaults to no data uncertainty.
        cov_cols : list, optional
            List of strings with the names of the columns of ``data`` that contain the
            data's covariance.
            Must have the same length as ``data_cols``.
            If ``cov`` is given but ``cov_cols`` is not, it defaults to appending
            ``'_cov'`` to the two respective entries of ``data_cols``.

        See Also
        --------
        :func:`~pandas.date_range` : Quick function to generate a timevector.
        """
        assert len(timevector) == data.shape[0], \
            "Length of 'timevector' has to match the number of rows in 'data', " \
            f"got {len(timevector)} and {data.shape}."
        assert isinstance(data_cols, list) and \
            all([isinstance(dcol, str) for dcol in data_cols]), \
            f"'data_cols' has to be a list of strings, got {data_cols}."
        # create dataframe with data
        df_data = {dcol: data[:, icol] for icol, dcol in enumerate(data_cols)}
        # add variances
        if var is None:
            var_cols = None
        else:
            assert data.shape == var.shape, \
                "'data' and 'var' need to have the same shape, got " \
                f"{data.shape} and {var.shape}."
            if var_cols is None:
                var_cols = [dcol + "_var" for dcol in data_cols]
            else:
                assert len(var_cols) == var.shape[1]
            df_data.update({vcol: var[:, icol] for icol, vcol in enumerate(var_cols)})
            # if variances are given, can also check for covariances
            if cov is None:
                cov_cols = None
            else:
                num_components = data.shape[1]
                cov_dims = get_cov_dims(num_components)
                if cov.ndim == 1:
                    cov = cov.reshape(-1, 1)
                assert cov.shape == (data.shape[0], cov_dims), \
                    "'data' and 'cov' need to have compatible shapes, got " \
                    f"{data.shape} and {cov.shape} (need {cov_dims} columns)."
                if cov_cols is None:
                    cov_cols = []
                    for i1 in range(num_components):
                        for i2 in range(i1 + 1, num_components):
                            cov_cols.append(f"{data_cols[i1]}_{data_cols[i2]}_cov")
                else:
                    assert len(cov_cols) == cov.shape[1]
                df_data.update({ccol: cov[:, icol] for icol, ccol in enumerate(cov_cols)})
        df = pd.DataFrame(data=df_data, index=timevector)
        return cls(df, src, data_unit, data_cols, var_cols, cov_cols)


class GipsyTimeseries(Timeseries):
    """
    Subclasses :class:`~Timeseries`.

    Timeseries subclass for GNSS measurements in JPL's Gipsy ``.tseries`` file format.
    The data and (co)variances are converted into millimeters (squared).

    Parameters
    ----------
    path : str
        Path to the timeseries file.
    show_warnings : bool, optional
        If ``True``, warn if there are data inconsistencies encountered while loading.
        Defaults to ``True``.
    data_unit : str, optional
        Can be ``'mm'`` (default) or ``'m'``.

    Notes
    -----

    The column format is described on `JPL's website`_:

    +---------------+-------------------------------------------------+
    | Columns       | Description                                     |
    +===============+=================================================+
    | Column 1      | Decimal year computed with 365.25 days/yr       |
    +---------------+-------------------------------------------------+
    | Columns 2-4   | East, North and Vertical [m]                    |
    +---------------+-------------------------------------------------+
    | Columns 5-7   | East, North and Vertical standard deviation [m] |
    +---------------+-------------------------------------------------+
    | Columns 8-10  | East, North and Vertical correlation [-]        |
    +---------------+-------------------------------------------------+
    | Column 11     | Time in Seconds past J2000                      |
    +---------------+-------------------------------------------------+
    | Columns 12-17 | Time in YEAR MM DD HR MN SS                     |
    +---------------+-------------------------------------------------+

    Time is GPS time, and the time series are relative to each station's first epoch.

    (Note that the documentation says covariance, but it is actually correlation,
    as per personal communication with Angelyn Moore, JPL.)

    .. _JPL's website: https://sideshow.jpl.nasa.gov/pub/JPL_GPS_Timeseries/\
repro2018a/raw/position/envseries/0000_README.format

    """
    def __init__(self, path, show_warnings=True, data_unit="mm"):
        self._path = str(path)
        if data_unit == "m":
            factor = 1
        elif data_unit == "mm":
            factor = 1000
        else:
            raise ValueError(f"'data_unit' needs to be 'mm' or 'm', got {data_unit}.")
        # load data
        data_cols = ["east", "north", "up"]
        var_cols = ["east_var", "north_var", "up_var"]
        cov_cols = ["east_north_cov", "east_up_cov", "north_up_cov"]
        all_cols = data_cols + var_cols + cov_cols
        time = pd.read_csv(self._path, delim_whitespace=True, header=None,
                           usecols=[11, 12, 13, 14, 15, 16],
                           names=["year", "month", "day", "hour", "minute", "second"])
        time = pd.to_datetime(time).to_frame(name="time")
        data = pd.read_csv(self._path, delim_whitespace=True, header=None,
                           usecols=[1, 2, 3, 4, 5, 6, 7, 8, 9],
                           names=all_cols)
        # compute covariance from correlation, still in meters
        data.loc[:, "east_north_cov"] *= data.loc[:, "east_var"] * data.loc[:, "north_var"]
        data.loc[:, "east_up_cov"] *= data.loc[:, "east_var"] * data.loc[:, "up_var"]
        data.loc[:, "north_up_cov"] *= data.loc[:, "north_var"] * data.loc[:, "up_var"]
        # convert standard deviation into variance, still in meters
        data.loc[:, var_cols] *= data.loc[:, var_cols]
        # now convert everything to millimeters
        data.loc[:, data_cols] *= factor
        data.loc[:, var_cols + cov_cols] *= factor**2
        # combine dataframe with time
        df = time.join(data)
        # check for duplicate timestamps and create time index
        num_duplicates = int(df.duplicated(subset="time").sum())
        if show_warnings and (num_duplicates > 0):
            warn(f"Timeseries file {self._path} contains data for {num_duplicates} "
                 "duplicate dates. Keeping first occurrences.", stacklevel=2)
        df = df.drop_duplicates(subset="time").set_index("time")
        # check for monotonic time index
        if not df.index.is_monotonic_increasing:
            warn(f"Timeseries file {self._path} is not ordered in time, sorting it now.",
                 stacklevel=2)
            df.sort_index(inplace=True)
        # construct Timeseries object
        super().__init__(dataframe=df, src=".tseries", data_unit=data_unit,
                         data_cols=data_cols, var_cols=var_cols, cov_cols=cov_cols)

    def get_arch(self):
        """
        Returns a JSON-compatible dictionary with all the information necessary to recreate
        the Timeseries instance (provided the data file is available).

        Returns
        -------
        dict
            JSON-compatible dictionary sufficient to recreate the GipsyTimeseries instance.

        See Also
        --------
        Timeseries.get_arch : For further information.
        """
        return {"type": "GipsyTimeseries",
                "kw_args": {"path": self._path}}


class UNRTimeseries(Timeseries):
    """
    Subclasses :class:`~Timeseries`.

    Timeseries subclass for GNSS measurements in UNR's ``.tenv3`` file format.
    The data and (co)variances are converted into millimeters (squared).

    Parameters
    ----------
    path : str
        Path to the timeseries file.
    show_warnings : bool, optional
        If ``True``, warn if there are data inconsistencies encountered while loading.
        Defaults to ``True``.
    data_unit : str, optional
        Can be ``'mm'`` (default) or ``'m'``.

    Notes
    -----

    The column format is described on `UNR's website`_:

    +---------------+---------------------------------------------------+
    | Columns       | Description                                       |
    +===============+===================================================+
    | Column 1      | Station name                                      |
    +---------------+---------------------------------------------------+
    | Column 2      | Date                                              |
    +---------------+---------------------------------------------------+
    | Column 3      | Decimal year                                      |
    +---------------+---------------------------------------------------+
    | Column 4      | Modified Julian day                               |
    +---------------+---------------------------------------------------+
    | Columns 5-6   | GPS week and day                                  |
    +---------------+---------------------------------------------------+
    | Column 7      | Longitude [°] of reference meridian               |
    +---------------+---------------------------------------------------+
    | Columns 8-9   | Easting [m] from ref. mer., integer and fraction  |
    +---------------+---------------------------------------------------+
    | Columns 10-11 | Northing [m] from equator, integer and fraction   |
    +---------------+---------------------------------------------------+
    | Columns 12-13 | Vertical [m], integer and fraction                |
    +---------------+---------------------------------------------------+
    | Column 14     | Antenna height [m]                                |
    +---------------+---------------------------------------------------+
    | Column 15-17  | East, North, Vertical standard deviation [m]      |
    +---------------+---------------------------------------------------+
    | Column 18     | East-North correlation coefficient [-]            |
    +---------------+---------------------------------------------------+
    | Column 19     | East-Vertical correlation coefficient [-]         |
    +---------------+---------------------------------------------------+
    | Column 20     | North-Vertical correlation coefficient [-]        |
    +---------------+---------------------------------------------------+

    Newer files also contain the following three columns:

    +---------------+---------------------------------------------------+
    | Column 21     | Latitude [°]                                      |
    +---------------+---------------------------------------------------+
    | Column 22     | Longitude [°]                                     |
    +---------------+---------------------------------------------------+
    | Column 23     | Altitude [m]                                      |
    +---------------+---------------------------------------------------+

    The time series are relative to each station's first integer epoch.

    .. _UNR's website: http://geodesy.unr.edu/gps_timeseries/README_tenv3.txt

    """
    def __init__(self, path, show_warnings=True, data_unit="mm"):
        self._path = str(path)
        if data_unit == "m":
            factor = 1
        elif data_unit == "mm":
            factor = 1000
        else:
            raise ValueError(f"'data_unit' needs to be 'mm' or 'm', got {data_unit}.")
        # load data and check for some warnings
        df = pd.read_csv(self._path, delim_whitespace=True,
                         usecols=[0, 3] + list(range(6, 13)) + list(range(14, 20)))
        if show_warnings and len(df['site'].unique()) > 1:
            warn(f"Timeseries file {self._path} contains multiple site codes: "
                 f"{df['site'].unique()}", stacklevel=2)
        if len(df['reflon'].unique()) > 1:
            raise NotImplementedError(f"Timeseries file {self._path} contains "
                                      "multiple reference longitudes: "
                                      f"{df['reflon'].unique()}")
        if len(df['_e0(m)'].unique()) > 1:
            if show_warnings:
                warn(f"Timeseries file {self._path} contains multiple integer "
                     f"Eastings: {df['_e0(m)'].unique()}", stacklevel=2)
            offsets_east = df["_e0(m)"].values - df["_e0(m)"].values[0]
        else:
            offsets_east = 0
        if len(df['____n0(m)'].unique()) > 1:
            if show_warnings:
                warn(f"Timeseries file {self._path} contains multiple integer "
                     f"Northings: {df['____n0(m)'].unique()}", stacklevel=2)
            offsets_north = df["____n0(m)"].values - df["____n0(m)"].values[0]
        else:
            offsets_north = 0
        if len(df['u0(m)'].unique()) > 1:
            if show_warnings:
                warn(f"Timeseries file {self._path} contains multiple integer "
                     f"Verticals: {df['u0(m)'].unique()}", stacklevel=2)
            offsets_up = df["u0(m)"].values - df["u0(m)"].values[0]
        else:
            offsets_up = 0
        # remove columns that are no longer needed
        df.drop(columns=["site", "reflon"], inplace=True)
        # make the data
        df["east"] = (df["__east(m)"] + offsets_east) * factor
        df["north"] = (df["_north(m)"] + offsets_north) * factor
        df["up"] = (df["____up(m)"] + offsets_up) * factor
        df.drop(columns=["_e0(m)", "__east(m)", "____n0(m)", "_north(m)",
                         "u0(m)", "____up(m)"], inplace=True)
        # make the covariance
        df["__corr_en"] *= df["sig_e(m)"] * df["sig_n(m)"] * factor**2
        df["__corr_eu"] *= df["sig_e(m)"] * df["sig_u(m)"] * factor**2
        df["__corr_nu"] *= df["sig_n(m)"] * df["sig_u(m)"] * factor**2
        df.rename(columns={"__corr_en": "east_north_cov", "__corr_eu": "east_up_cov",
                           "__corr_nu": "north_up_cov"}, inplace=True)
        # make the variance
        old_sig_cols = ["sig_e(m)", "sig_n(m)", "sig_u(m)"]
        df.loc[:, old_sig_cols] = (df.loc[:, old_sig_cols] * factor)**2
        df.rename(columns={"sig_e(m)": "east_var", "sig_n(m)": "north_var",
                           "sig_u(m)": "up_var"}, inplace=True)
        # check for duplicate timestamps and create time index
        num_duplicates = int(df.duplicated(subset="__MJD").sum())
        if show_warnings and (num_duplicates > 0):
            warn(f"Timeseries file {self._path} contains data for {num_duplicates} "
                 "duplicate dates. Keeping first occurrences.", stacklevel=2)
        df.drop_duplicates(subset="__MJD", inplace=True)
        df["__MJD"] = pd.to_datetime(df["__MJD"] + 2400000.5, unit="D", origin='julian')
        df.rename(columns={"__MJD": "time"}, inplace=True)
        df.set_index("time", inplace=True)
        # check for monotonic time index
        if not df.index.is_monotonic_increasing:
            warn(f"Timeseries file {self._path} is not ordered in time, sorting it now.",
                 stacklevel=2)
            df.sort_index(inplace=True)
        # construct Timeseries object
        super().__init__(dataframe=df, src=".tenv3", data_unit=data_unit,
                         data_cols=["east", "north", "up"],
                         var_cols=["east_var", "north_var", "up_var"],
                         cov_cols=["east_north_cov", "east_up_cov", "north_up_cov"])

    def get_arch(self):
        """
        Returns a JSON-compatible dictionary with all the information necessary to recreate
        the Timeseries instance (provided the data file is available).

        Returns
        -------
        dict
            JSON-compatible dictionary sufficient to recreate the UNRTimeseries instance.

        See Also
        --------
        Timeseries.get_arch : For further information.
        """
        return {"type": "UNRTimeseries",
                "kw_args": {"path": self._path}}
